package explicit;

import java.util.BitSet;
import java.util.Vector;

import cy95.LTLManager;
import parser.ast.Expression;
import parser.ast.ExpressionLabel;
import parser.ast.ExpressionTemporal;
import parser.ast.ExpressionUnaryOp;
import parser.type.TypeDouble;
import prism.PrismComponent;
import prism.PrismException;

public class DTMCModelCheckerCY95 extends DTMCModelChecker {

	// public static class Partition {
	// 	BitSet almostSure;
	// 	BitSet maybe;
	// 	BitSet almostNever;
	// 	
	// 	public void swapAlmostSureAlmostNever() {
	// 		BitSet newAlmostNever = almostSure;
	// 		almostSure = almostNever;
	// 		almostNever = newAlmostNever;
	// 	}
	// 	
	// 	/**
	// 	 * Restricts every almostSure, almostNever and maybe to reach
	// 	 * <br>[ REFS: <i>none</i>, DEREFS: reach ]
	// 	 * @param reach
	// 	 */
	// 	public void restrict(BitSet reach)
	// 	{
	// 		almostSure.and(reach);
	// 		almostNever.and(reach);
	// 		maybe.and(reach);
	// 	}
	// }
	
	protected DTMCExplicit model;
	
	public DTMCModelCheckerCY95(PrismComponent parent) throws PrismException {
		super(parent);
	}
	
	@Override
	protected StateValues checkProbPathFormulaLTL(Model model, Expression expr, boolean qual, MinMax minMax, BitSet statesOfInterest) throws PrismException
	{
		return checkProbPathFormulaLTL(model, expr, qual, minMax, statesOfInterest, model.getNumStates());
	}
	
	protected StateValues checkProbPathFormulaLTL(Model model, Expression expr, boolean qual, MinMax minMax, BitSet statesOfInterest, int numberOfStates) throws PrismException
	{
		this.model = (DTMCExplicit)model;
		Vector<BitSet> labelBS = new Vector<BitSet>();

		// Take over already known labels
		for(String label : model.getLabels()) {
			if(label.startsWith("L")) {
				try {
					int index = Integer.parseInt(label.substring(1));
					labelBS.add(index, model.getLabelStates(label));
				} catch(Exception e) {
					// Do nothing
				}
			}
		}
		LTLModelChecker ltlMC = new LTLModelChecker(this);
		// Model check maximal state formulas
		Expression ltl = ltlMC.checkMaximalStateFormulas(this, model, expr.deepCopy(), labelBS);
		
		// Add new labels to model
		for(int i = 0; i < labelBS.size(); i++) {
			if(!model.hasLabel("L" + i)) {
				this.model.addLabel("L" + i, labelBS.elementAt(i));
			}
		}

		ExpressionTemporal innerTemp = LTLManager.getInnerTemporal(ltl);
		if(innerTemp == null) {
			// Remove unnecessary parentheses
			expr = LTLManager.removeParenthesis(expr);
			mainLog.println("Transformed model has " + model.getNumStates() + " states");

			if(Expression.isTrue(expr)) {
				return new StateValues(TypeDouble.getInstance(), Double.valueOf(1.0), model);
			}
			
			if(Expression.isFalse(expr)) {
				return new StateValues(TypeDouble.getInstance(), Double.valueOf(0.0), model);
			}
			
			// innerTemp must be a label
			ExpressionLabel exprLabel = (ExpressionLabel)expr;
			if(model instanceof DTMCSparseCY95) {
				DTMCSparseCY95 dtmcCY95 = (DTMCSparseCY95)model;
				BitSet goalStates = dtmcCY95.getLabelStates(exprLabel.getName());
				StateValues result = new StateValues(TypeDouble.getInstance(), numberOfStates);
				for(int i = goalStates.nextSetBit(0); i >= 0; i = goalStates.nextSetBit(i + 1)) {
					if(dtmcCY95.getInitDistr().containsKey(i)) {
						int originalState = dtmcCY95.getOriginalStates().get(i);
						result.getDoubleArray()[originalState] += dtmcCY95.getInitDistr().get(i);
					}
				}
				return result;
			} else {
				StateValues result = new StateValues(TypeDouble.getInstance(), numberOfStates);
				BitSet goalStates = model.getLabelStates(exprLabel.getName());
				for(int i = goalStates.nextSetBit(0); i >= 0; i = goalStates.nextSetBit(i + 1)) {
					result.getDoubleArray()[i] = 1;
				}
				return result;
			}
		}
		
		ExpressionLabel freshAtom = getSkolemLabel();
		DTMCSparseCY95 transformedDTMC = new DTMCSparseCY95(this.model, innerTemp, this);
		
		Expression newLTL = LTLManager.replaceTemporal(ltl, innerTemp, freshAtom);
		
		// mainLog.println("Model for " + newLTL);
		// transformedDTMC.exportToDotFile(mainLog);
		transformedDTMC.addSkolemLabel(freshAtom);
		
		DTMCModelCheckerCY95 newCY95 = new DTMCModelCheckerCY95(this);
		return newCY95.checkProbPathFormulaLTL(transformedDTMC, newLTL, qual, minMax, statesOfInterest, numberOfStates);
	}

	protected ExpressionLabel getSkolemLabel()
	{
		int index = 0;
		for(String labelString : model.getLabels()) {
			if(labelString.matches("Skolem\\d+")) {
				try {
					index = Math.max(index, Integer.parseInt(labelString.substring(6)) + 1);
				} catch(Exception e) {
					
				}
			}
		}
		
		return new ExpressionLabel("Skolem" + index);
	}
}
