package explicit;

import java.util.AbstractMap;
import java.util.Arrays;
import java.util.BitSet;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.TreeMap;

import parser.ast.Expression;
import parser.ast.ExpressionLabel;
import parser.ast.ExpressionTemporal;
import parser.ast.ExpressionUnaryOp;
import prism.PrismException;

public class DTMCSparseCY95 extends DTMCSparse {
	
	protected DTMCModelChecker mc;
	protected DTMC dtmc;

	protected Map<Integer, Map<Boolean, Integer>> oldState2newState = new TreeMap<>();
	protected int currentStateIndex;
	protected int column = 0;
	
	protected Map<Integer, Double> initDistr = new HashMap<>();
	protected Set<ExpressionLabel> skolemLabels; 
	
	protected double[] values;
	protected Map<Integer, Integer> originalStates;
	
	protected Map<Integer, Entry<int[], double[]>> unwrittenAdjencency = new TreeMap<>();
	protected int lastWrittenAdjencency = -1;

	public DTMCSparseCY95(DTMCExplicit dtmc, ExpressionTemporal expr, DTMCModelChecker mc) throws PrismException
	{
		super();
		this.mc = mc;
		this.dtmc = dtmc;
		constantValues = dtmc.getConstantValues();
		varList = dtmc.getVarList();
		if(ExpressionTemporal.isNext(expr)) {
			// Get Bitset of Label
			BitSet goal = expression2BitSet(expr.getOperand2());
			transformNext(goal);
		} else {
			if(ExpressionTemporal.isFinally(expr)) {
				// Get Bitset of Label
				BitSet goal = expression2BitSet(expr.getOperand2());
				transformFinally(goal);
			} else if (ExpressionTemporal.isGlobally(expr)) {
				// Get Bitset of Label
				BitSet goal = expression2BitSet(expr.getOperand2());

				transformGlobally(goal);
			} else if (ExpressionTemporal.isUntil(expr)) {
				BitSet left = expression2BitSet(expr.getOperand1());
				BitSet right = expression2BitSet(expr.getOperand2());
				
				transformUntil(left, right);
			} else if (ExpressionTemporal.isRelease(expr)) {
				BitSet left = expression2BitSet(expr.getOperand1());
				BitSet right = expression2BitSet(expr.getOperand2());
				
				transformRelease(left, right);
			} else if(ExpressionTemporal.isWeakUntil(expr)) {
				BitSet left = expression2BitSet(expr.getOperand1());
				BitSet right = expression2BitSet(expr.getOperand2());
				
				transformWeak(left, right);
			}
		}
		
		flushCache();
		
		createInitDistr();
		takeOverLabels();
		mapToOriginalStates(dtmc);
	}
	
	protected void mapToOriginalStates(DTMCExplicit dtmc) {
		originalStates = new HashMap<>();
		for(Entry<Integer, Map<Boolean, Integer>> entry : oldState2newState.entrySet()) {
			for(int transformedState : entry.getValue().values()) {
				int originalState;
				if(dtmc instanceof DTMCSparseCY95) {
					originalState = ((DTMCSparseCY95) dtmc).getOriginalStates().get(entry.getKey());
				} else {
					originalState = entry.getKey();
				}
				originalStates.put(transformedState, 
						originalState);
			}
		}

		if(dtmc instanceof DTMCSparseCY95) {
			((DTMCSparseCY95)dtmc).getOriginalStates().clear();
		}		
	}

	protected Map<Integer, Integer> getOriginalStates()
	{
		return originalStates;
	}

	private void createInitDistr()
	{
		if(dtmc instanceof DTMCSparseCY95) {
			Map<Integer, Double> oldInitDistr = ((DTMCSparseCY95)dtmc).getInitDistr();
			for(Entry<Integer, Double> initState : oldInitDistr.entrySet()) {
				if(values[initState.getKey()] == 0.0 || values[initState.getKey()] == 1.0) {
					initDistr.put(
							oldState2newState.get(initState.getKey()).get(values[initState.getKey()] == 1.0),
							oldInitDistr.get(initState.getKey())
							);
				} else {
					initDistr.put(
							oldState2newState.get(initState.getKey()).get(true), 
							oldInitDistr.get(initState.getKey()) * values[initState.getKey()]
							);
					initDistr.put(
							oldState2newState.get(initState.getKey()).get(false), 
							oldInitDistr.get(initState.getKey()) * (1 - values[initState.getKey()])
							);
				}
			}
			oldInitDistr.clear();
		} else {
			int initState = dtmc.getFirstInitialState();
			if(values[initState] == 0.0 || values[initState] == 1.0) {
				initDistr.put(
						oldState2newState.get(initState).get(values[initState] == 1.0),
						1.0);
			} else {
				initDistr.put(
						oldState2newState.get(initState).get(true), 
						values[initState]);
				initDistr.put(
						oldState2newState.get(initState).get(false), 
						1 - values[initState]);
			}
		}
	}
	
	protected void takeOverLabels()
	{
		for(String oldLabel : dtmc.getLabels()) {
			BitSet label = dtmc.getLabelStates(oldLabel);
			BitSet newLabel = new BitSet(getNumStates());
			for(int i = label.nextSetBit(0); i >= 0; i = label.nextSetBit(i + 1)) {
				if(oldState2newState.get(i).containsKey(true)) {
					newLabel.set(oldState2newState.get(i).get(true));
				}
				if(oldState2newState.get(i).containsKey(false)) {
					newLabel.set(oldState2newState.get(i).get(false));
				}
			}
			addLabel(oldLabel, newLabel);
		}
		
		if(dtmc instanceof DTMCSparseCY95) {
			skolemLabels = ((DTMCSparseCY95)dtmc).getSkolemLabels();
		} else {
			skolemLabels = new HashSet<>();
		}
	}
	
	public Set<ExpressionLabel> getSkolemLabels()
	{
		return skolemLabels;
	}

	public Map<Integer, Double> getInitDistr()
	{
		return initDistr;
	}

	protected void transformUntil(BitSet left, BitSet right) throws PrismException
	{
		// Calculate Probability vector
		values = mc.computeUntilProbs(dtmc, left, right).soln;
		initialise();

		// Transform according algorithm
		for(int oldState = 0; oldState < dtmc.getNumStates(); oldState++) {
			double value = values[oldState];
			if(value == 0.0 || value == 1.0) {
				transformUntilAlmostSureNever(oldState);
			}  else {
				transformUntilMaybe(oldState);
			}
		}
	}
	
	protected void initialise()
	{
		long stateNumAlmostSureNever = Arrays.stream(values).filter(x -> (x==0.0 || x==1.0)).count();
		numStates = (int) (2*values.length - stateNumAlmostSureNever);
		rows = new int[numStates + 1];
		int numTransitions = 0;
		for(int state = 0; state < dtmc.getNumStates(); state++) {
			Iterator<Integer> succIt = dtmc.getSuccessorsIterator(state);
			while(succIt.hasNext()) {
				int succ = succIt.next();
				numTransitions += (values[succ] != 0.0 && values[succ] != 1.0) ? 2 : 1;
			}
		}
		
		rows[numStates] = numTransitions;
		columns = new int[numTransitions];
		probabilities = new double[numTransitions];
	}

	protected void transformUntilMaybe(int oldStartState)
	{
		int newStartStateSatisfied = findOrAdd(oldState2newState, oldStartState, true);
		int newStartStateUnsatisfied = findOrAdd(oldState2newState, oldStartState, false);

		int indexSatisfied = 0;
		int indexUnsatisfied = 0;

		int succsSatisfied[] = new int[2*dtmc.getNumTransitions(oldStartState)];
		double probsSatisfied[] = new double[2*dtmc.getNumTransitions(oldStartState)];

		int succsUnsatisfied[] = new int[2*dtmc.getNumTransitions(oldStartState)];
		double probsUnsatisfied[] = new double[2*dtmc.getNumTransitions(oldStartState)];

		Iterator<Entry<Integer, Double>> it = dtmc.getTransitionsIterator(oldStartState);
		while(it.hasNext()) {
			Entry<Integer, Double> trans = it.next();
			int succ = trans.getKey();
			
			// For successor with P(a U b) = 0 just take transition probabilities over
			if(values[succ] == 0.0) {
				int succNew = findOrAdd(oldState2newState, succ, false);

				succsUnsatisfied[indexUnsatisfied] = succNew;
				probsUnsatisfied[indexUnsatisfied] = trans.getValue() / (1 - values[oldStartState]);
				indexUnsatisfied++;
			} else
			// For successor with P(a U b) = 1 just take transition probabilities over
			if(values[succ] == 1.0) {
				int succNew = findOrAdd(oldState2newState, succ, true);

				succsSatisfied[indexSatisfied] = succNew;
				probsSatisfied[indexSatisfied] = trans.getValue() / values[oldStartState];
				indexSatisfied++;
			} else {
				// For successors with 0 < P(a U b) < 1 take transitions 
				int succNewSatisfied = findOrAdd(oldState2newState, succ, true);
				int succNewUnsatisfied = findOrAdd(oldState2newState, succ, false);
			
				succsUnsatisfied[indexUnsatisfied] = succNewUnsatisfied;
				probsUnsatisfied[indexUnsatisfied] = trans.getValue() * (1 - values[succ]) / (1 - values[oldStartState]);
				indexUnsatisfied++;

				succsSatisfied[indexSatisfied] = succNewSatisfied;
				probsSatisfied[indexSatisfied] = trans.getValue() * values[succ] / values[oldStartState];
				indexSatisfied++;
			}
		}
		
		setProbability(newStartStateSatisfied, succsSatisfied, probsSatisfied);
		setProbability(newStartStateUnsatisfied, succsUnsatisfied, probsUnsatisfied);
		
	}

	protected void transformUntilAlmostSureNever(int oldStartState)
	{
		int newStartState = findOrAdd(oldState2newState, oldStartState, values[oldStartState] == 1.0);
		int index = 0;
		int succs[] = new int[2*dtmc.getNumTransitions(oldStartState)];
		double probs[] = new double[2*dtmc.getNumTransitions(oldStartState)];
		Iterator<Entry<Integer, Double>> it = dtmc.getTransitionsIterator(oldStartState);
		while(it.hasNext()) {
			Entry<Integer, Double> trans = it.next();
			int succ = trans.getKey();
			
			// For successor with P(a U b) = 0 just take transition probabilities over
			if(values[succ] == 0.0 || values[succ] == 1.0) {
				int succNew = findOrAdd(oldState2newState, succ, values[succ] == 1.0);
				succs[index] = succNew;
				probs[index] = trans.getValue();
				index++;
			} else {
				int succNewSatisfied = findOrAdd(oldState2newState, succ, true);
				int succNewUnsatisfied = findOrAdd(oldState2newState, succ, false);
			
				succs[index] = succNewSatisfied;
				probs[index] = trans.getValue() * values[succ];
				index++;

				succs[index] = succNewUnsatisfied;
				probs[index] = trans.getValue() * (1-values[succ]);
				index++;
			}
		}
		
		setProbability(newStartState, succs, probs);
		
	}

	protected int findOrAdd(Map<Integer, Map<Boolean, Integer>> oldState2newState, int oldState, boolean satisfied)
	{
		if(!oldState2newState.containsKey(oldState)) {
			Map<Boolean, Integer> newState = new HashMap<>();
			int result = currentStateIndex;
			currentStateIndex++;
			newState.put(satisfied, result);
			oldState2newState.put(oldState, newState);
			return result;
		} else {
			Map<Boolean, Integer> newState = oldState2newState.get(oldState);
			if(newState.containsKey(satisfied)) {
				return newState.get(satisfied);
			} else {
				int result = currentStateIndex;
				currentStateIndex++;
				newState.put(satisfied, result);
				return result;
			}
		}
	}

	protected void transformNextAlmostSureNever(int oldStartState)
	{
		int newStartState = findOrAdd(oldState2newState, oldStartState, values[oldStartState] == 1.0);
		int index = 0;
		int succs[] = new int[2*dtmc.getNumTransitions(oldStartState)];
		double probs[] = new double[2*dtmc.getNumTransitions(oldStartState)];
		Iterator<Entry<Integer, Double>> it = dtmc.getTransitionsIterator(oldStartState);
		while(it.hasNext()) {
			Entry<Integer, Double> trans = it.next();
			int succ = trans.getKey();
			
			// For successor with P(X a) = 0 just take transition probabilities over
			if(values[succ] == 0.0 || values[succ] == 1.0) {
				int succNew = findOrAdd(oldState2newState, succ, values[succ] == 1.0);
				succs[index] = succNew;
				probs[index] = trans.getValue();
				index++;
			} else {
				int succNewSatisfied = findOrAdd(oldState2newState, succ, true);
				int succNewUnsatisfied = findOrAdd(oldState2newState, succ, false);
			
				succs[index] = succNewSatisfied;
				probs[index] = trans.getValue() * values[succ];
				index++;
				succs[index] = succNewUnsatisfied;
				probs[index] = trans.getValue() * (1 - values[succ]);
				index++;
			}
		}

		setProbability(newStartState, succs, probs);
	}

	protected void transformNextMaybe(int oldStartState, BitSet goal)
	{
		int newStartStateSatisfied = findOrAdd(oldState2newState, oldStartState, true);
		int newStartStateUnsatisfied = findOrAdd(oldState2newState, oldStartState, false);

		int indexSatisfied = 0;
		int indexUnsatisfied = 0;

		int succsSatisfied[] = new int[2*dtmc.getNumTransitions(oldStartState)];
		double probsSatisfied[] = new double[2*dtmc.getNumTransitions(oldStartState)];

		int succsUnsatisfied[] = new int[2*dtmc.getNumTransitions(oldStartState)];
		double probsUnsatisfied[] = new double[2*dtmc.getNumTransitions(oldStartState)];

		Iterator<Entry<Integer, Double>> it = dtmc.getTransitionsIterator(oldStartState);
		while(it.hasNext()) {
			Entry<Integer, Double> trans = it.next();
			int succ = trans.getKey();
			
			// For successor with P(a U b) = 0 just take transition probabilities over
			if(values[succ] == 0.0 || values[succ] == 1.0) {
				int succNew = findOrAdd(oldState2newState,
						succ,
						values[succ] == 1.0);
				if(goal.get(trans.getKey())) {
					succsSatisfied[indexSatisfied] = succNew;
					probsSatisfied[indexSatisfied] = trans.getValue() / values[oldStartState];
					indexSatisfied++;
				} else {
					succsUnsatisfied[indexUnsatisfied] = succNew;
					probsUnsatisfied[indexUnsatisfied] = trans.getValue() / (1 - values[oldStartState]);
					indexUnsatisfied++;
				}
			} else { 
				int succNewSatisfied = findOrAdd(oldState2newState, succ, true);
				int succNewUnsatisfied = findOrAdd(oldState2newState, succ, false);
			
				if(goal.get(trans.getKey())) {
					succsSatisfied[indexSatisfied] = succNewSatisfied;
					probsSatisfied[indexSatisfied] = trans.getValue() * values[succ] / values[oldStartState];
					indexSatisfied++;

					succsSatisfied[indexSatisfied] = succNewUnsatisfied;
					probsSatisfied[indexSatisfied] = trans.getValue() * (1 - values[succ]) / values[oldStartState];
					indexSatisfied++;
				} else {
					succsUnsatisfied[indexUnsatisfied] = succNewSatisfied;
					probsUnsatisfied[indexUnsatisfied] = trans.getValue() * values[succ] / (1 - values[oldStartState]);
					indexUnsatisfied++;

					succsUnsatisfied[indexUnsatisfied] = succNewUnsatisfied;
					probsUnsatisfied[indexUnsatisfied] = trans.getValue() * (1 - values[succ]) / (1 - values[oldStartState]);
					indexUnsatisfied++;
				}
			}
		}

		setProbability(newStartStateSatisfied, succsSatisfied, probsSatisfied);
		setProbability(newStartStateUnsatisfied, succsUnsatisfied, probsUnsatisfied);
	}
		
	protected void transformNext(BitSet goal) throws PrismException
	{
		// Calculate Probability vector
		values = mc.computeNextProbs(dtmc, goal).soln;
		initialise();
		
		// Transform according algorithm
		for(int oldState = 0; oldState < dtmc.getNumStates(); oldState++) {
			double value = values[oldState];
			if(value == 0.0 || value == 1.0) {
				transformNextAlmostSureNever(oldState);
			} else {
				transformNextMaybe(oldState, goal);
			}
		}
	}

	protected void transformFinally(BitSet goal) throws PrismException
	{
		// Calculate Probability vector
		values = mc.computeReachProbs(dtmc, goal).soln;
		initialise();

		for(int oldState = 0; oldState < dtmc.getNumStates(); oldState++) {
			double value = values[oldState];
			if(value == 0.0 || value == 1.0) {
				transformUntilAlmostSureNever(oldState);
			} else {
				transformUntilMaybe(oldState);
			}
		}
	}

	protected void transformGlobally(BitSet goal) throws PrismException
	{
		// G a = ! F ! a
		goal.flip(0, dtmc.getNumStates());
		// Calculate Probability vector
		values = mc.computeReachProbs(dtmc, goal).soln;
		for(int i = 0; i < values.length; i++) {
			values[i] = 1.0 - values[i];
		}
		initialise();
			
		for(int oldState = 0; oldState < dtmc.getNumStates(); oldState++) {
			double value = values[oldState];
			if(value == 0.0 || value == 1.0) {
				transformUntilAlmostSureNever(oldState);
			} else {
				transformUntilMaybe(oldState);
			}
		}
	}

	protected void transformWeak(BitSet left, BitSet right) throws PrismException
	{
		// a W b = !((a & !b) U (!a & !b))
		BitSet rightCompl = right;
		rightCompl.flip(0, dtmc.getNumStates());
		
		BitSet leftCompl = (BitSet)left.clone();
		leftCompl.flip(0, dtmc.getNumStates());
		
		BitSet leftUntil = left;
		leftUntil.and(rightCompl);
		
		BitSet rightUntil = rightCompl;
		rightCompl.and(leftCompl);
		
		// Calculate Probability vector
		values = mc.computeUntilProbs(dtmc, leftUntil, rightUntil).soln;
		for(int i = 0; i < values.length; i++) {
			values[i] = 1.0 - values[i];
		}
		initialise();

		// Transform according algorithm
		for(int oldState = 0; oldState < dtmc.getNumStates(); oldState++) {
			double value = values[oldState];
			if(value == 0.0 || value == 1.0) {
				transformUntilAlmostSureNever(oldState);
			} else {
				transformUntilMaybe(oldState);
			}
		}
	}

	protected void transformRelease(BitSet left, BitSet right) throws PrismException
	{
		// a R b = ! (!a U !b)
		left.flip(0, dtmc.getNumStates());
		right.flip(0, dtmc.getNumStates());
		
		// Calculate Probability vector
		values = mc.computeUntilProbs(dtmc, left, right).soln;
		for(int i = 0; i < values.length; i++) {
			values[i] = 1.0 - values[i];
		}
		initialise();

		for(int oldState = 0; oldState < dtmc.getNumStates(); oldState++) {
			double value = values[oldState];
			if(value == 0.0 || value == 1.0) {
				transformUntilAlmostSureNever(oldState);
			} else {
				transformUntilMaybe(oldState);
			}
		}
	}
	
	/**
	 * Sets the transition probabilities for all transitions outgoing of a particular state
	 * @param start the starting state of the transitions
	 * @param goal a vector with indices of the transition ends
	 * @param prob a vector with the probabilities of the according vector
	 */
	protected void setProbability(int start, int goal[], double prob[])
	{
		if(start == lastWrittenAdjencency + 1) {
			rows[start] = column;
			for(int index = 0; index < goal.length; index++) {
				if(prob[index] > 0) {
					columns[column] = goal[index];
					probabilities[column] = prob[index];
					column++;
				}
			}
			
			lastWrittenAdjencency = start;
			
			while(unwrittenAdjencency.containsKey(lastWrittenAdjencency + 1)) {
				lastWrittenAdjencency++;
				flushCache(lastWrittenAdjencency);
			}
		} else {
			unwrittenAdjencency.put(start, new AbstractMap.SimpleEntry<>(goal, prob));
		}
	}
	
	protected void flushCache()
	{
		for(int i = lastWrittenAdjencency + 1; i < currentStateIndex; i++) {
			rows[i] = column;
			if(unwrittenAdjencency.containsKey(i)) {
				flushCache(i);
			}
		}
		
		unwrittenAdjencency.clear();
	}
	
	protected void flushCache(int start) {
		rows[start] = column;
		int[] goal = unwrittenAdjencency.get(start).getKey();
		double[] prob = unwrittenAdjencency.get(start).getValue();
		for(int index = 0; index < goal.length; index++) {
			if(prob[index] > 0) {
				columns[column] = goal[index];
				probabilities[column] = prob[index];
				column++;
			}
		}
		unwrittenAdjencency.remove(start);
	}
	
	public void addSkolemLabel(ExpressionLabel skolemLabel)
	{
		BitSet skolemStates = new BitSet(getNumStates());
		for(int i = 0; i < dtmc.getNumStates(); i++) {
			if(oldState2newState.containsKey(i) && oldState2newState.get(i).containsKey(true)) {
				skolemStates.set(oldState2newState.get(i).get(true));
			}
		}
		addLabel(skolemLabel.getName(), skolemStates);
		skolemLabels.add(skolemLabel);
	}
	
	private BitSet expression2BitSet(Expression expr) {
		Expression inner = expr;
		boolean negated = false;

		// Get innermost label
		while(Expression.isNot(inner) || Expression.isParenth(inner)) {
			ExpressionUnaryOp unary = (ExpressionUnaryOp)inner;
			if(Expression.isNot(inner)) {
				negated = !negated;
			}
			inner = unary.getOperand();
		}
		
		if(Expression.isTrue(inner)) {
			BitSet goal = new BitSet(dtmc.getNumStates());
			goal.set(0, dtmc.getNumStates());
			return goal;
		}
		
		if(Expression.isFalse(inner)) {
			BitSet goal = new BitSet(dtmc.getNumStates());
			goal.set(0, dtmc.getNumStates(), false);
			return goal;
			
		}
		
		ExpressionLabel goalLabel = (ExpressionLabel)inner;
		BitSet goal = dtmc.getLabelStates(goalLabel.getName());
		
		if(negated) {
			goal.flip(0, dtmc.getNumStates());
		}
		
		// Get Bitset of Label
		return goal;
	}
	
	// private boolean rowsInsane() {
	// 	for(int i = 0; i < rows.length - 1; i++) {
	// 		if(rows[i] > rows[i+1] && rows[i+1] != 0) {
	// 			return true;
	// 		}
	// 	}
	// 	return false;
	// }
}
