package prism;

import jdd.JDD;
import jdd.JDDNode;
import jdd.JDDVars;
import parser.ast.Expression;
import parser.ast.ExpressionLabel;
import parser.ast.ExpressionTemporal;

public class CY95TransformatorNext extends CY95Transformator {
	
	protected JDDNode goal;

	public CY95TransformatorNext(ProbModel model, ProbModelCheckerCY95 mc, ExpressionTemporal temporal, boolean qual,
			JDDNode statesOfInterest) throws PrismException {
		super(model, mc, temporal, qual, statesOfInterest);

		ExpressionLabel goalLabel = (ExpressionLabel) (temporal.getOperand2());
		if(originalModel.hasLabelDD(goalLabel.getName())) {
			goal = originalModel.getLabelDD(goalLabel.getName());
		} else {
			int index = Integer.parseInt(goalLabel.getName().substring(1));
			goal = labelDDs.get(index);
		}

		reachValues = mc.computeNextProbs(model.getTrans(), goal).convertToStateValuesMTBDD();
	}

	@Override
	protected Partition divide(ExpressionTemporal temporal, boolean qual, JDDNode statesOfInterest)
	{
		if(Expression.isTrue(temporal.getOperand2())) {
			return divideTrue();
		}

		// Unregarded by the temporal operator, if the right/the inner temporal formula is true
		// Everything is true
		JDDNode goal = labelToDD(temporal.getOperand2()).copy();
	
		if(ExpressionTemporal.isNext(temporal)) {
			return divideNext(goal.copy());
		}
		
		return null;
	}

	public JDDNode getTransformedTrans()
	{
		JDDNode newTrans = super.getTransformedTrans();
		
		// Case start state is maybe state
		// Calculate normalised probability, i.e., prob(goal satisfied)/prob(start satisfied)
		JDDNode goalPrimed = JDD.PermuteVariables(goal.copy(), originalModel.getAllDDRowVars(), originalModel.getAllDDColVars());
		JDDNode transGoalSatisfied = JDD.Times(
				partition.maybe.copy(),
				newMaybeStatesPrimed.copy(),
				rowSkolemVar.copy(),
				normalize(originalModel.getTrans().copy(), maybeValuesDD.copy(), false),
				goalPrimed.copy());
		newTrans = JDD.Plus(newTrans, 
				JDD.Times(
					transGoalSatisfied.copy(),
					maybeValuesDDPrimed.copy(),
					colSkolemVar.copy()
					)
				);
		newTrans = JDD.Plus(newTrans, 
				JDD.Times(
					transGoalSatisfied.copy(), 
					JDD.Apply(JDD.MINUS, JDD.ONE.copy(), maybeValuesDDPrimed.copy()),
					JDD.Not(colSkolemVar.copy())
					)
				);

		JDDNode transGoalNotSatisfied = JDD.Times(
				partition.maybe.copy(),
				newMaybeStatesPrimed.copy(),
				JDD.Not(rowSkolemVar.copy()),
				normalize(originalModel.getTrans().copy(), maybeValuesDD.copy(), true),
				JDD.Not(goalPrimed.copy()));
		
		newTrans = JDD.Plus(newTrans, 
				JDD.Times(
					transGoalNotSatisfied.copy(),
					maybeValuesDDPrimed.copy(),
					colSkolemVar.copy()
					)
				);
		newTrans = JDD.Plus(newTrans, 
				JDD.Times(
					transGoalNotSatisfied.copy(), 
					JDD.Apply(JDD.MINUS, JDD.ONE.copy(), maybeValuesDDPrimed.copy()),
					JDD.Not(colSkolemVar.copy())
					)
				);
		
		newTrans = JDD.Plus(newTrans,
				JDD.Times(
						partition.maybe.copy(),
						rowSkolemVar.copy(),
						normalize(originalModel.getTrans().copy(), maybeValuesDD.copy(), false),
						JDD.Or(newOneStatesPrimed.copy(), newZeroStatesPrimed.copy()),
						goalPrimed.copy()
						)
				);
		newTrans = JDD.Plus(newTrans,
				JDD.Times(
						partition.maybe.copy(),
						JDD.Not(rowSkolemVar.copy()),
						normalize(originalModel.getTrans().copy(), maybeValuesDD.copy(), true),
						JDD.Or(newOneStatesPrimed.copy(), newZeroStatesPrimed.copy()),
						JDD.Not(goalPrimed)
					)
				);

		return newTrans;
	}

	protected Partition divideNext(JDDNode goal)
	{		
		Partition partition = new Partition();
		JDDVars allDDRowVars = originalModel.getAllDDRowVars();
		JDDVars allDDColVars = originalModel.getAllDDColVars();
		JDDNode preNotGoal = directPre(JDD.Not(goal.copy()), allDDRowVars, allDDColVars);
		partition.almostSure = JDD.Not(preNotGoal.copy());
		partition.maybe = JDD.And(directPre(goal, allDDRowVars, allDDColVars), preNotGoal);
		partition.almostNever = JDD.Not(JDD.Or(partition.almostSure.copy(), partition.maybe.copy()));
		
		return partition;
		
	}
	
}
